;  
;                                                                      
;   Filename  :   CLOCKCLK.ASM - System Clock Cloaking Device          
;   Version   :   1.0                                                  
;   Author    :   Thomas W. Olsen, CIS: [76450,1767]                       
;   Assembler :   Microsoft Macro Assembler (MASM) v6.0B               
;   Build Note:   ml /Zm clockclk.asm                                  
;                                                                      
;  

        .386P
        include CLOAKAPI.INC        

;
; Constant & Structure Definitions
;

        TICKS_PER_SECOND EQU    18
        TRUE             EQU    1
        FALSE            EQU    0
        CR               EQU    13
        LF               EQU    10
        VIDEO_SEGMENT    EQU    0B800h
        DISPLAY_COLUMN   EQU    75 * 2
        BIOS_SEGMENT     EQU    0
        TIMER_OFFSET     EQU    46Ch

        XMSMoveStructure struc
                BlockLength     dw      0,0
                SourceHandle    dw      0       ;0=Conventional Memory
                SourceOffset    dw      0,0
                DestHandle      dw      0       ;0=Conventional Memory
                DestOffset      dw      0,0     ;FlatOffset or Seg:Offset if Conventional
        XMSMoveStructure ends

;
; V86 Entrypoint Code 
;

V86_CODE Segment Public Use16 Byte 'V86'
        Assume cs:V86_CODE, ds:V86_CODE, ss:STACK

V86ResidentCodeBegin label byte

        OldTimerOffset  dw      ?               
        OldTimerSegment dw      ?
        Reentrant       dw      FALSE
        Ticks           db      TICKS_PER_SECOND

V86_TimerHandler proc far
        pushf
        call    dword ptr cs:OldTimerOffset     ;Call Original Timer Tick
        cmp     cs:Reentrant, TRUE              ;Are We Becoming Reentrant?
        je      @@0
        dec     cs:Ticks                        
        jz      @@1                             ;Has One Second Expired?
@@0:
        iret                                    ;Abort If Reentrant
@@1:
        mov     cs:Reentrant, TRUE              ;Set Reentrant Flag
        mov     cs:Ticks, TICKS_PER_SECOND      ;Reset Counter to One Second
        int     2CH                             ;RM386 Transition to PM Code
@@2:
        sti                                     ;Enable Interrupts
        mov     cs:Reentrant, FALSE             ;Clear Reentrant Flag
        retf    2                               ;Return Flags
V86_TimerHandler endp

V86ResidentCodeEnd label byte

        Psp                     dw      0
        BaseSegment             dw      0
        XmsHandle               dw      0
        XmsAddress              dd      0
        CloakingDeviceLoaded    db      CR,LF,"System Clock Cloaking Device Loaded.",CR,LF,"$"
        ErrorRM386NotLoaded     db      CR,LF,"RM386 Not loaded.",CR,LF,"$"
        ErrorXMSNotLoaded       db      CR,LF,"XMS Memory Manager Not Loaded.",CR,LF,"$"
        ErrorOutOfMemory        db      CR,LF,"Insufficient Extended Memory.",CR,LF,"$"
        ErrorCannotLockMemory   db      CR,LF,"Cannot Lock Memory Region.",CR,LF,"$"
        ErrorCannotRelocateCode db      CR,LF,"Cannot Load Protected Mode Code.",CR,LF,"$"
        ErrorCannotExecCode     db      CR,LF,"Cannot Activate Protected Mode Code.",CR,LF,"$"
        XMSMoveStruc            XMSMoveStructure <>

V86_Main proc    near
        public  V86_Main

        mov     ax, 351Ch
        int     21h                     ;DOS Get Interrupt Vector 1Ch
        mov     cs:OldTimerOffset, bx
        mov     cs:OldTimerSegment, es  ;ES:BX -> Old 1Ch Vector
        
        mov     cs:Psp, ds              ;Save PSP Segment
        mov     cs:BaseSegment, ds
        add     cs:BaseSegment, 6       ;Move V86_TimerHandler Code Into PSP:60h
        mov     ax, cs                  ; to Minimize Our Memory Footprint Even More
        mov     ds, ax
        mov     si, offset V86ResidentCodeBegin
        mov     es, cs:BaseSegment
        xor     di, di
        mov     cx, offset V86ResidentCodeEnd
        sub     cx, offset V86ResidentCodeBegin
        cld
        rep     movsb                   ;Copy Code into PSP (Overlays FCB Area)

        mov     ax, cs
        mov     es, ax

        mov     ax,3567h                ;Is EMS Manager Present?
        int     21h                     
        mov     ax, es
        or      ax, bx                   
        jne     @@3
        mov     dx, offset ErrorRM386NotLoaded
        jmp     @@10
@@3:    
        mov     ax,5BF0h                ;Is RM386 Manager Present?
        int     67h
        or      ah,ah
        je      @@4
        mov     dx,offset ErrorRM386NotLoaded
        jmp     @@10
@@4:    
        mov     ax,4300h                ;Is XMS Manager Present?
        int     2fh
        cmp     al,80h
        je      @@5
        mov     dx,offset ErrorXMSNotLoaded
        jmp     @@10
@@5:    
        mov     ax,4310h                ;Get XMS Manager Address
        int     2Fh
        mov     word ptr cs:XmsAddress,   bx
        mov     word ptr cs:XmsAddress+2, es

        mov     ax, offset PMCodeEnd    ;Compute Amt of XMS Memory to Hold PM Code
        sub     ax, offset PMCodeBegin         
        shr     ax, 10                  ;Convert Bytes -> Kilobytes
        inc     ax                      ;Round Up
        mov     dx,ax

        mov     ah,9                    ;XMS Allocate Memory
        call    dword ptr XmsAddress
        or      ax,ax                   ;Successful?
        jne     @@6
        mov     dx,offset ErrorOutOfMemory
        jmp     @@10
@@6:    
        mov     XmsHandle,dx            ;Save Handle
        mov     ah, 0ch                 ;XMS Lock Memory Region
        call    dword ptr XmsAddress    
        or      ax, ax                  ;Successful?
        jne     @@7
        mov     dx,offset ErrorCannotLockMemory
        jmp     @@10
@@7:    
        mov     ax, XmsHandle           ;Move PM Code into XMS Alloc'd Block        
        mov     XMSMoveStruc.DestHandle, ax
        xor     eax, eax
        mov     ax,  offset PMCodeEnd           
        xor     ebx, ebx
        mov     bx,  offset PMCodeBegin
        sub     eax, ebx                ;BlockLength = PMCodeEnd - PMCodeBegin
        shr     eax, 4                  ;Convert to Paragraphs
        inc     eax                     ;Round Up
        shl     eax, 4                  ;Convert Back to Bytes

        mov     dword ptr XMSMoveStruc.Blocklength, eax
        mov     ax, PM_CODE
        mov     XMSMoveStruc.SourceOffset+2,ax
        mov     si,offset XMSMoveStruc
        mov     ah,0bh
        call    dword ptr XmsAddress
        or      ax,ax
        jne     @@8
        mov     dx,offset ErrorCannotRelocateCode
        jmp     @@10
@@8:                                    
        mov     si, cs:BaseSegment      ;ESI Uniquely Identifies Our V86 Stub
        shl     esi, 16
        mov     si, offset @@2          ;Pass INT 2CH Iret Address to PM Init
        mov     dx, XmsHandle
        mov     cl, 1                   ;32-Bit Code
        mov     ah, 82h                 ;RM386 Transition to PM Init Code
        call    dword ptr XmsAddress    
        or      ax, ax                  ;Successful?
        jne     @@9
        mov     dx, offset ErrorCannotExecCode
        jmp     @@10
@@9:                                    ;PM Code is Running...Grab V86 Ints and TSR
        mov     dx, offset V86_TimerHandler
        mov     ax, cs:BaseSegment      ;V86_TimerHandler Code Seg Now Equals PSP+6h
        mov     ds, ax
        mov     ax, 251Ch
        int     21h                     ;DOS Set Interrupt Vector 1Ch

        mov     es, cs:Psp              ;Point to PSP
        mov     es, es:[2CH]            ;Get Environment Segment
        mov     ah, 49h                 ;Deallocate Environment, Too.
        int     21h                     

        mov     ax, cs
        mov     ds, ax
        lea     dx, CloakingDeviceLoaded
        mov     ah, 9
        int     21h

        mov     dx, cs:BaseSegment
        sub     dx, cs:Psp
        shl     dx, 4
        add     dx, offset V86ResidentCodeEnd
        shr     dx, 4
        inc     dx
        mov     ax,3100h                ;Terminate & Stay Resident
        int     21h
@@10:  
        mov     ax,cs                   ;Error Occurred ... Cleanup Time
        mov     ds,ax
        push    dx                      
        mov     dx,XmsHandle            ;Was any XMS Allocated?
        or      dx,dx
        je      @@11                    ;Yes ... Free It Before Terminating
        mov     ah,0ah
        call    dword ptr cs:XmsAddress

@@11:   pop     dx                      ;DOS Display Error Message
        mov     ah,9
        int     21h
        mov     ax,4C00h                ;Terminate.
        int     21h

V86_Main endp

V86_CODE ENDS

;
; Protected Mode Code
;

PM_CODE Segment Public Use32 Para 'PMCODE'
        Assume  cs:PM_CODE, ds:PM_CODE

PMCodeBegin label byte

        jmp     PM_Init

        PMDataSelector  dw      0
        V86DataSelector dw      0
        FlatSelector    dw      0
        PassAlong       df      0
        RM386_API       df      0
        VideoSelector   dw      0
        V86Address      dd      0

        Hours           dw      ?
                        db     ':'
        Minutes         dw      ?
                        db      0

PM_Init proc near
        public  PM_Init
        mov     PMDataSelector, ds
        mov     V86DataSelector, es
        mov     FlatSelector, gs
        mov     V86Address, esi

        mov     ax, get_passalong       ;Get Head of the PassAlong DaisyChain
        int     2CH
        mov     dword ptr ds:PassAlong, ebx
        mov     word  ptr ds:PassAlong+4, dx

        mov     ebx, offset PM_TimerHandler
        mov     dx, cs
        mov     ax, set_passalong       ;Make Ourselves the Head of DaisyChain
        int     2CH
        retf
PM_Init endp



PM_TimerHandler PROC FAR
        public  PM_TimerHandler
        cmp     eax, cs:V86Address      ;Were we Called by V86_TimerHandler?
        je      @@12
        cmp     word ptr cs:[PassAlong+4],0
        je      @@13                    ;Are there Any Other Passalongs?
        jmp     fword ptr cs:PassAlong  ;Yes ... DaisyChain to Next
@@12:   
        call    PM_ShowClock            ;Display the Clock
        stc
@@13:
        ret
PM_TimerHandler ENDP


PM_ShowClock proc near
        public  PM_ShowClock
        mov     ds, cs:PMDataSelector
        mov     es, cs:V86DataSelector
        xor     edi, edi
        mov     di,  BIOS_SEGMENT       ;Get Master Clock Count from 
        shl     edi, 4
        add     edi, TIMER_OFFSET       ; V86 BIOS Data Area
        mov     eax, es:[edi]           ; V86 0:[46Ch] --> PM V86DataSelector:[0*10h+46Ch]
        xor     edx, edx
        mov     ecx, 65544
        div     ecx
        push    eax                     ;EAX = Hour, EDX = Remainder
        xchg    eax, edx
        xor     edx, edx
        mov     ecx, 1092
        div     ecx                     ;EAX = Minutes, EDX = Remainder
        push    eax

        pop     eax
        call    PM_HexToAscii           ;Input:  AX = Unsigned Hex Integer
        mov     Minutes, ax             ;Output: AX = 2-Byte ASCII Representation

        pop     eax
        call    PM_HexToAscii           ;Input:  AX = Unsigned Hex Integer
        mov     Hours, ax               ;Output: AX = 2-Byte ASCII Representation

        mov     esi, offset Hours       ;Display Time Directly to Video Buffer
        xor     edi, edi
        mov     di, VIDEO_SEGMENT
        shl     edi, 4
        add     edi, DISPLAY_COLUMN     ; V86 0:[B800h] --> 
        cld                             ;   PM V86DataSelector:[B800h*10h+DISPLAY_COLUMN*2]
        mov     ah, es:[edi+1]
@@14:   lodsb                           ;Load Character from PM Data Area
        and     al, al                  ;End of String?
        jz      @@15
        stosw                           ;Move Character to V86 Video Buffer
        jmp     @@14
@@15:
        ret
PM_ShowClock endp

PM_HexToAscii proc near
        push    ecx
        push    edx
        xor     edx, edx
        xor     ah, ah
        mov     cx, 10
        div     cx
        shl     al, 4
        or      al, dl
        shl     ax, 4
        xchg    ah, al
        shr     ah, 4
        add     ax, 3030h
        pop     edx
        pop     ecx
        ret
PM_HexToAscii endp

PMCodeEnd label byte

PM_CODE Ends


STACK   Segment Stack 'STACK'
                dw      512 DUP(0)
STACK   Ends

        End     V86_Main
